/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - libcschem (core library)
 *  Copyright (C) 2022 Tibor 'Igor2' Palinkas
 *
 *  (Supported by NLnet NGI0 PET Fund in 2022)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */


#include <libcschem/config.h>

#include <assert.h>

#include <genvector/gds_char.h>

#include <libcschem/concrete.h>
#include <libcschem/cnc_any_obj.h>
#include <libcschem/plug_library.h>
#include <libcschem/event.h>

#include <librnd/core/compat_misc.h>
#include <librnd/core/error.h>
#include <librnd/core/plugins.h>

#include <sch-rnd/conf_core.h>

static csch_lib_backend_t be_symlib_local;
static csch_lib_master_t *loclib_master;

static const char symlib_local_cookie[] = "symlib_local";

static unsigned ll_hash(const void *obj)                   { return csch_cobj_hash((void *)obj); }
static int ll_keyeq(const void *obj1, const void *obj2)    { return csch_cobj_keyeq((void *)obj1, (void *)obj2); }
static unsigned ll_hashna(const void *obj)                 { return csch_cobj_hash_((void *)obj, 1); }
static int ll_keyeqna(const void *obj1, const void *obj2)  { return csch_cobj_keyeq_((void *)obj1, (void *)obj2, 1); }

/* hash: a htpi_t, keyed with indirect symlib groups are stored in the
   indirect/purpose=symlib group's backend_data->ptr[0]; [1] is the version
   without attribute matching. Also remember our local root group in ptr[2].
*/
static void symlib_local_sheet_init(rnd_hidlib_t *hl, csch_lib_t *root_dir, const csch_cgrp_t *symlib)
{
	root_dir->backend_data.ptr[0] = htpi_alloc(ll_hash, ll_keyeq);
	root_dir->backend_data.ptr[1] = htpi_alloc(ll_hashna, ll_keyeqna);
	root_dir->backend_data.ptr[2] = (void *)symlib;
}

static void symlib_local_sheet_uninit(csch_lib_t *root_dir)
{
	csch_cgrp_t *symlib = root_dir->backend_data.ptr[2];

	if (symlib == NULL) return;

	htpi_free(root_dir->backend_data.ptr[0]);
	htpi_free(root_dir->backend_data.ptr[1]);
	root_dir->backend_data.ptr[0] = NULL;
	root_dir->backend_data.ptr[1] = NULL;
	root_dir->backend_data.ptr[2] = NULL;
}

static int symlib_local_map_local_(rnd_hidlib_t *hl, csch_lib_t *root_dir, csch_cgrp_t *symlib)
{
	htip_entry_t *e;
	htpi_t *sh = root_dir->backend_data.ptr[0];
	htpi_t *shna = root_dir->backend_data.ptr[1];

	for(e = htip_first(&symlib->id2obj); e != NULL; e = htip_next(&symlib->id2obj, e)) {
		csch_cgrp_t *grp = e->value;
		if ((grp->hdr.type == CSCH_CTYPE_GRP) && (grp->role == CSCH_ROLE_SYMBOL)) {
			const char *gname = csch_attrib_get_str(&grp->attr, "name");
			csch_lib_t *newent;
			char tmp[64];
			gds_t nm = {0};

			if (gname != NULL)
				gds_append_str(&nm, gname);
			sprintf(tmp, ".%d",  grp->hdr.oid);
			gds_append_str(&nm, tmp);

			if (htpi_has(sh, grp)) {
				rnd_message(RND_MSG_WARNING, "Redundant local lib symbol '%s'\n", nm.array);
			}
			else {
				htpi_set(sh, grp, 0);
				htpi_set(shna, grp, 0);
			}

			newent = csch_lib_alloc_append(&be_symlib_local, root_dir, nm.array, CSCH_SLIB_STATIC);
			newent->backend_data.lng[0] = grp->hdr.oid;
			/* do not free nm: name ownership passed on to newent */
		}
	}
	return 0;
}

static csch_cgrp_t *get_sheet_local_symlib(const csch_cgrp_t *indirect, int alloc)
{
	htip_entry_t *e;
	for(e = htip_first(&indirect->id2obj); e != NULL; e = htip_next(&indirect->id2obj, e)) {
		csch_cgrp_t *grp = e->value;
		const char *purp = csch_attrib_get_str(&grp->attr, "purpose");
		if ((purp != NULL) && (strcmp(purp, "symlib") == 0))
			return grp;
	}
	if (alloc) {
		TODO("alloc");
		abort();
	}
	return NULL;
}

static int symlib_local_map_local(rnd_hidlib_t *hl, csch_lib_t *root_dir, const csch_cgrp_t *indirect)
{
	csch_cgrp_t *grp = get_sheet_local_symlib(indirect, 0);
	if (grp != NULL) {
		symlib_local_sheet_init(hl, root_dir, grp);
		symlib_local_map_local_(hl, root_dir, grp);
	}
	return 0;
}


static int symlib_local_load(csch_sheet_t *sheet, void *dst, csch_lib_t *src, const char *params)
{
	if (params != NULL)
		return -1;
	return -1;
}

static void symlib_local_free(csch_lib_t *src)
{
	symlib_local_sheet_uninit(src->parent);
}

/* We have *grp already placed on the sheet; move it to indirect; this happens
   if some other plugin already handled the event and decided to place the
   group */
static void local_paste_already_placed(csch_sheet_t *sheet, csch_chdr_t **grp)
{
	/* no other plugin handles this event at the moment */
	rnd_message(RND_MSG_ERROR, "symlib_local local_paste_already_placed() not yet implemented\nPlease report this bug!\n");
}

/* insert *grp in indirect, place ref on the sheet */
static void local_paste_place(csch_sheet_t *sheet, csch_chdr_t **grp)
{
	csch_chdr_t *loc;
	csch_cgrp_t *symlib = get_sheet_local_symlib(&sheet->indirect, 1);
	csch_lib_root_t *libroot;
	csch_lib_t *root_dir;
	htpi_t *sh;
	htpi_entry_t *e;

	if (loclib_master->uid >= sheet->local_libs.used) {
		rnd_message(RND_MSG_ERROR, "local_paste_place(): out of bounds #1\nPlease report this bug!\n");
		/* csch_lib_add_all(sheet, loclib_master, NULL) */
		return;
	}

	libroot = sheet->local_libs.array[loclib_master->uid];
	if (sheet->local_libs.used < 1) {
		rnd_message(RND_MSG_ERROR, "local_paste_place(): out of bounds #1\nPlease report this bug!\n");
		return;
	}

	root_dir = libroot->roots.array[0];
	sh = root_dir->backend_data.ptr[0];

	if (sh == NULL) { /* local lib is not yet created */
		symlib_local_sheet_init(&sheet->hidlib, root_dir, symlib);
		sh = root_dir->backend_data.ptr[0];
		assert(sh != NULL);
	}

	/* check for direct match: the very same symbol already in lib */
	e = htpi_getentry(sh, *grp);
	rnd_trace("*** loclib: root=%p sh=%p e=%p\n", root_dir, sh, e);

	if (0) { /* hash debug: take a sheet with exactly 1 local lib entry and place it */
		htip_entry_t *e = htpi_first(sh);
		csch_chdr_t *l = e->key;
		long lh, gh;
		int eq;

		rnd_trace("---- h l ----\n");
		lh = csch_cobj_hash_(l, 0);
		rnd_trace("---- h g ----\n");
		gh = csch_cobj_hash_(*grp, 0);
		rnd_trace("---- eq ----\n");
		eq = csch_cobj_keyeq_(l, *grp, 0);
		rnd_trace("l=%p lh=%ld gh=%ld -> %d\n", l, lh, gh, eq);
	}

	TODO("if enabled: place in indirect, replace *obj with a ref");
}


/* Perform a "place a ref from local lib" when pasting a symbol from the buffer */
static void symlib_local_paste(rnd_hidlib_t *hidlib, void *user_data, int argc, rnd_event_arg_t argv[])
{
	csch_sheet_t *sheet = (csch_sheet_t *)hidlib;
	csch_chdr_t **obj = argv[1].d.p;

	if (!conf_core.editor.paste_to_local_lib)
		return;

	if (argv[1].type != RND_EVARG_PTR) {
		rnd_message(RND_MSG_ERROR, "symlib_local_paste() wrong first arg\n");
		return;
	}

	if ((*obj)->type == CSCH_CTYPE_GRP) {
		csch_cgrp_t *grp = (csch_cgrp_t *)*obj;
		if ((grp->role == CSCH_ROLE_SYMBOL) && (grp->sym_prefer_loclib)) {
			if (grp->hdr.sheet == sheet)
				local_paste_already_placed(sheet, obj);
			else
				local_paste_place(sheet, obj);
		}
	}
}


int pplg_check_ver_symlib_local(int ver_needed) { return 0; }

void pplg_uninit_symlib_local(void)
{
	rnd_event_unbind_allcookie(symlib_local_cookie);
}

int pplg_init_symlib_local(void)
{
	RND_API_CHK_VER;

	be_symlib_local.name = "symlib_local";
	be_symlib_local.map_local = symlib_local_map_local;
	be_symlib_local.load = symlib_local_load;
	be_symlib_local.free = symlib_local_free;

	loclib_master = csch_lib_get_master("symbol", 1);
	csch_lib_backend_reg(loclib_master, &be_symlib_local);

	rnd_event_bind(CSCH_EVENT_BUFFER_PASTE_CUSTOM, symlib_local_paste, NULL, symlib_local_cookie);

	return 0;
}

